use crossbeam_channel::unbounded;
use socket2::SockAddr;
use tokio::{net::UdpSocket, sync::Mutex};
use base64::prelude::*;
use log::{error, info, warn};
use std::sync::Arc;
use std::net::Ipv4Addr;
use x25519_dalek::{PublicKey, StaticSecret};
use std::{
    fs::File,
    io::{self, Read, Write, BufReader, BufWriter},
    net::SocketAddr,
    process::Command,
    os::unix::io::FromRawFd,
};
use aes_gcm::{
    aead::{Aead, AeadCore, KeyInit, OsRng},
    Aes256Gcm, Nonce};

use crate::config::ClientConfiguration;
use crate::udp::{UDPVpnPacket, UDPVpnHandshake, UDPSerializable};
use network_interface::NetworkInterface;
use network_interface::NetworkInterfaceConfig;
use robusta_jni::jni::JNIEnv;
use crate::jni::FridaLib;

pub async fn client_mode(client_config: ClientConfiguration, fd: i32, env: &JNIEnv<'_>) {
    let (ltx, lrx) = unbounded::<Vec<u8>>();
    ltx.send("Starting client...".as_bytes().to_vec());

    let sock = UdpSocket::bind("0.0.0.0:25565").await.unwrap();
    sock.connect(&client_config.server.endpoint).await.unwrap();
   
    ltx.send(format!("FD: {:?}", &fd).as_bytes().to_vec());
    ltx.send("So exhausting".as_bytes().to_vec());
    let mut dev = unsafe { File::from_raw_fd(fd) };
    let mut dev1 = dev.try_clone().unwrap();
    let mut dev_reader = BufReader::new(dev);
    let mut dev_writer = BufWriter::new(dev1);

    let sock_rec = Arc::new(sock);
    let sock_snd = sock_rec.clone();

    let (tx, rx) = unbounded::<Vec<u8>>();
    let (dx, mx) = unbounded::<Vec<u8>>();

    let cipher_shared: Arc<Mutex<Option<x25519_dalek::SharedSecret>>> = Arc::new(Mutex::new(None));

    tokio::spawn(async move {
        while let Ok(bytes) = rx.recv() {
            ltx.send("Write to tun.".as_bytes().to_vec());
            dev_writer.write_all(&bytes).unwrap();
        }
    });

    tokio::spawn(async move {
        let mut buf = vec![0; 1400]; // mtu
        while let Ok(n) = dev_reader.read(&mut buf) {
            ltx.send("Read from tun.".as_bytes().to_vec());
            dx.send(buf[..n].to_vec()).unwrap();
        }
    });

    let priv_key = BASE64_STANDARD.decode(client_config.client.private_key).unwrap();
    
    let cipher_shared_clone = cipher_shared.clone();
    tokio::spawn(async move {
        let mut buf = vec![0; 4096];

        loop {
            if let Ok(l) = sock_rec.recv(&mut buf).await {
                ltx.send("Read from socket".as_bytes().to_vec());
                let mut s_cipher = cipher_shared_clone.lock().await;
                match buf.first() {
                    Some(h) => {
                        match h {
                            0 => {
                                let handshake = UDPVpnHandshake::deserialize(&(buf[..l].to_vec()));
                                let mut k = [0u8; 32];
                                for (&x, p) in handshake.public_key.iter().zip(k.iter_mut()) {
                                    *p = x;
                                }
                                let mut k1 = [0u8; 32];
                                for (&x, p) in priv_key.iter().zip(k1.iter_mut()) {
                                    *p = x;
                                }
                                *s_cipher = Some(StaticSecret::from(k1)
                                    .diffie_hellman(&PublicKey::from(k)));
                            }, // handshake
                            1 => {
                                let wrapped_packet = UDPVpnPacket::deserialize(&(buf[..l].to_vec()));
                                if s_cipher.is_some() {
                                    let aes = Aes256Gcm::new(s_cipher.as_ref().unwrap().as_bytes().into());
                                    let nonce = Nonce::clone_from_slice(&wrapped_packet.nonce);
                                    match aes.decrypt(&nonce, &wrapped_packet.data[..]) {
                                        Ok(decrypted) => { let _ = tx.send(decrypted); },
                                        Err(error) => { ltx.send(format!("Decryption error! {:?}", error).as_bytes().to_vec()); }
                                    }
                                } else {
                                    ltx.send("There is no static_secret".as_bytes().to_vec());
                                }
                            }, // payload
                            2 => { ltx.send("Got keepalive packet".as_bytes().to_vec()); },
                            _ => { ltx.send("Unexpected header value.".as_bytes().to_vec()); }
                        }
                    },
                    None => { ltx.send("There is no header.".as_bytes().to_vec()); }
                }
                drop(s_cipher);
            }
        }
    });

    let pkey = BASE64_STANDARD.decode(client_config.client.public_key).unwrap();
    let handshake = UDPVpnHandshake{ public_key: pkey, request_ip: client_config.client.address.parse::<Ipv4Addr>().unwrap() };
    let mut nz = 0;
    while nz < 25 {
        sock_snd.send(&handshake.serialize()).await.unwrap();
        nz += 1
    }
    //sock_snd.send(&handshake.serialize()).await.unwrap();

    let s_cipher = cipher_shared.clone();
    loop {
        tokio::select! {
            b = lrx.recv() => {
                FridaLib::traceFromNative(&env, String::from_utf8_lossy(&b.unwrap()).to_string());
            }
            b = mx.recv() => {
                let bytes = b.unwrap();
                let s_c = s_cipher.lock().await;
                
                if s_c.is_some() {
                    let aes = Aes256Gcm::new(s_c.as_ref().unwrap().as_bytes().into());
                    let nonce = Aes256Gcm::generate_nonce(&mut OsRng);
                    let ciphered_data = aes.encrypt(&nonce, &bytes[..]);
                    
                    if let Ok(ciphered_d) = ciphered_data {
                        let vpn_packet = UDPVpnPacket{ data: ciphered_d, nonce: nonce.to_vec()};
                        let serialized_data = vpn_packet.serialize();
                        ltx.send("Write to socket".as_bytes().to_vec());
                        sock_snd.send(&serialized_data).await.unwrap();
                    } else {
                        ltx.send("Socket encryption failed.".as_bytes().to_vec());
                    }
                } else {
                    ltx.send("There is no shared_secret in main loop".as_bytes().to_vec());
                }
            }
        }
    }
}